#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu
"""
(c) 2015 Ronan Delacroix
Job Manager Common sub module
:author: Ronan Delacroix
"""
import pkgutil
__path__ = pkgutil.extend_path(__path__, __name__)
from datetime import datetime
import os
import logging
import tempfile
import shutil
import mongoengine
import mongoengine.signals
import tbx
import tbx.process
import tbx.service
import tbx.log
import tbx.text
import uuid as UUID
import traceback
from datetime import datetime, timedelta


class ConfigurationException(Exception):
    pass


def update_modified(sender, document, **kwargs):
    document.updated = datetime.utcnow()


def public_dict(d):
    """ if type(d) is dict:
        return dict((k, public_dict(v)) for k, v in d.items() if not k.startswith('_'))
    elif type(d) is list:
        return [public_dict(v) for v in d]
    else:
        return d
    """

    if isinstance(d, dict):
        safe_dict = dict((key, public_dict(value)) for key, value in d.items() if not key.startswith('_'))
        if '_cls' in d:
            safe_dict['type'] = d['_cls']
        return safe_dict
    if isinstance(d, list):
        return [public_dict(l) for l in d]

    return d


class SerializableQuerySet(mongoengine.QuerySet):

    def to_json(self):
        return tbx.text.render_json(self.as_pymongo())

    def to_safe_dict(self):
        return [f.to_safe_dict() for f in self]
        #return [public_dict(f) for f in self.as_pymongo()]


def safely_import_from_name(modules):
    if modules:
        logging.info('Starting initial module import')
        logging.info('Trying to import the following : ' + ', '.join(modules))
    else:
        logging.warning('No modules to import.')
    for mod in modules:
        try:
            globals()[mod] = __import__(mod)
        except ImportError as e:
            logging.error("Can't import Job module '%s' as defined in settings. Exiting." % mod)
            raise ConfigurationException(e)
    if modules:
        logging.info('Modules import OK.')


class BaseDocument(mongoengine.Document):

    created = mongoengine.DateTimeField(required=True, default=datetime.utcnow)
    updated = mongoengine.DateTimeField(default=datetime.utcnow)

    meta = {
        'ordering': ['+created'],
        'allow_inheritance': True,
        'queryset_class': SerializableQuerySet,
        'abstract': True,
        'strict': False,
        'indexes': [
            'created',
        ]
    }

    def __init__(self, *args, **values):
        super(BaseDocument, self).__init__(*args, **values)

    @property
    def type(self):
        return self.__class__.__name__

    def to_json(self):
        return tbx.text.render_json(self.to_mongo())

    def to_safe_dict(self):
        return public_dict(self.to_mongo())

    def get_hash(self):
        import base64
        import hashlib
        return base64.b64encode(hashlib.sha1(mongoengine.Document.to_json(self, sort_keys=True).encode()).digest()).decode().strip('=').replace("+", "-")


class NamedDocument(BaseDocument):

    uuid = mongoengine.StringField(required=True, default=tbx.text.random_short_slug, unique=True)
    name = mongoengine.StringField(required=True)
    module = mongoengine.StringField()

    meta = {
        'ordering': ['+created'],
        'allow_inheritance': True,
        'queryset_class': SerializableQuerySet,
        'abstract': True,
        'indexes': [
            'uuid',
            'created',
        ]
    }

    def __init__(self, *args, **values):
        super(NamedDocument, self).__init__(*args, **values)
        if not self.name:
            self.name = self.type + ' ' + self.uuid

    def __repr__(self):
        return self.name


class LogProxy(object):

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def extra_log_arguments(self):
        return {}

    def log(self, text, method=logging.info):
        return method("%s - %s" % (self, text), extra=self.extra_log_arguments)

    def log_debug(self, text):
        return self.log(text, logging.debug)

    def log_info(self, text):
        return self.log(text, logging.info)

    def log_warning(self, text):
        return self.log(text, logging.warning)

    def log_error(self, text):
        return self.log(text, logging.error)

    def log_exception(self, text):
        return self.log(text, logging.exception)


class TempFolderProxy(object):

    @property
    def temp_folders(self):
        try:
            return self.__temp_folders
        except AttributeError:
            self.__temp_folders = []
        return self.__temp_folders

    def get_new_temp_folder(self, prefix=None):
        if prefix:
            prefix = prefix + '_'
        temp_folder = tempfile.mkdtemp(prefix=prefix)
        self.temp_folders.append(temp_folder)
        return temp_folder

    def get_named_temp_folder(self, name):
        temp_folder = os.path.join(tempfile.gettempdir(), name)
        os.makedirs(temp_folder, exist_ok=True)
        self.temp_folders.append(temp_folder)
        return temp_folder

    def clean_temp(self):

        if isinstance(self, LogProxy):
            log_func = self.log_debug
        else:
            def do_nothing(*args, **kwargs):
                pass
            log_func = do_nothing

        if not self.temp_folders:
            return

        log_func("Cleaning temporary folders.")
        for temp_folder in self.temp_folders:
            shutil.rmtree(temp_folder, ignore_errors=True)
            log_func("  - Removed %s" % temp_folder)
        self.__temp_folders = []
        log_func("Cleaning done.")


class Runnable(TempFolderProxy):

    status = mongoengine.StringField(required=True, default="pending", choices=('new', 'pending', 'running', 'success', 'error'))
    started = mongoengine.DateTimeField()
    finished = mongoengine.DateTimeField()
    details = mongoengine.StringField(required=False)
    result = mongoengine.DynamicField(required=False)

    def process(self, *args, **kwargs):
        raise NotImplementedError('The "process" method shall be subclassed to define the runnable processing.')

    def pre_process(self, *args, **kwargs):
        # should be overridden
        pass

    def post_process(self, result):
        self.result = result
        return self.result

    def safe_run(self, *args, **kwargs):
        kwargs['safe_run'] = True
        return self.run(*args, **kwargs)

    def run(self, *args, **kwargs):
        result = None
        safe_run = kwargs.pop('safe_run', False)
        self.started = datetime.utcnow()
        self.status = 'running'
        self.save()
        try:
            self.log_debug("Launching process...")
            self.pre_process(*args, **kwargs)
            result = self.process(*args, **kwargs)
            result = self.post_process(result)  # strangely can be useful
        except Exception as e:
            self.log_exception(e)
            self.details = "Exception : %s" % str(traceback.format_exc())
            self.status = 'error'
            self.finished = datetime.utcnow()
            self.save_as_error()
            if not safe_run:
                raise e
        else:
            self.status = 'success'
            self.finished = datetime.utcnow()
            self.save_as_successful()
        finally:
            self.log_info("Process ended (%s)." % self.status)
            self.clean_temp()
        return result

    def save_as_successful(self):
        self.save()

    def save_as_error(self):
        self.save()


class AutoDocumentable(object):
    @classmethod
    def get_doc(cls):
        import textwrap
        def create_documentation(cls):
            result = {
                "class": cls.__name__,
                "doc": textwrap.dedent(cls.__doc__) if cls.__doc__ else '',
            }

            if cls._subclasses:
                subclasses = list(cls._subclasses)
                if cls.__name__ in subclasses:
                    subclasses.remove(cls.__name__)
                if ('Job.' + cls.__name__) in subclasses:
                    subclasses.remove('Job.' + cls.__name__)

                if subclasses:
                    result['sub_classes'] = subclasses

            if cls._meta.get('abstract', False):
                result['abstract'] = True
                return result

            fields = {}
            for key, field in cls._fields.items():
                if key in ['id', '_cls']:
                    continue
                field_dict = {
                    'name': field.name,
                    'required': field.required
                }
                if isinstance(field, mongoengine.EmbeddedDocumentField):
                    field_doc = create_documentation(field.document_type_obj)
                    field_dict['type'] = field_doc['class']
                    if field_doc['doc']:
                        field_dict['type_doc'] = field_doc['doc']
                    if 'fields' in field_doc:
                        field_dict['fields'] = field_doc['fields']
                    if 'abstract' in field_doc:
                        field_dict['abstract'] = field_doc['abstract']
                    if 'sub_classes' in field_doc:
                        field_dict['sub_classes'] = field_doc['sub_classes']
                elif isinstance(field, mongoengine.EmbeddedDocumentListField):
                    field_doc = create_documentation(field.field.document_type_obj)
                    field_dict['type'] = 'List'
                    field_dict['list_type'] = field_doc['class']
                    if 'fields' in field_doc:
                        field_dict['list_type_fields'] = field_doc['fields']
                    if 'abstract' in field_doc:
                        field_dict['abstract'] = field_doc['abstract']
                    if 'sub_classes' in field_doc:
                        field_dict['sub_classes'] = field_doc['sub_classes']
                elif isinstance(field, mongoengine.ListField):
                    field_dict['type'] = 'List'
                    field_dict['list_type'] = field.field.__class__.__name__.replace('Field', '')
                else:
                    field_dict['type'] = field.__class__.__name__.replace('Field', '')

                if hasattr(field, 'doc') and field.doc:
                    field_dict['doc'] = field.doc

                if hasattr(field, 'default') and field.default:
                    if callable(field.default):
                        field_dict['default'] = str(field.default.__name__+'()')
                    else:
                        field_dict['default'] = str(field.default)

                if hasattr(field, 'min_value') and field.min_value:
                    field_dict['min_value'] = field.min_value
                if hasattr(field, 'max_value') and field.max_value:
                    field_dict['max_value'] = field.max_value
                if hasattr(field, 'choices') and field.choices:
                    field_dict['choices'] = list(field.choices)
                fields[field.name] = field_dict

            result["fields"] = fields
            return result

        return create_documentation(cls)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu
"""
(c) 2015-2017 Ronan Delacroix
Job Manager Host
:author: Ronan Delacroix
"""
import os
import sys
import logging
import socket
import tbx.code
from datetime import datetime, timedelta
import platform
import socket
import psutil
import pip
import mongoengine
import jobmanager
import jobmanager.common as common


class Host(common.BaseDocument):
    meta = {
        'ordering': ['-updated'],
        'queryset_class': common.SerializableQuerySet,
        'indexes': [
            'created',
            'updated',
            'hostname'
        ]
    }
    hostname = mongoengine.StringField(required=True)
    pid = mongoengine.IntField(required=True)
    mac_address = mongoengine.StringField()
    job_slots = mongoengine.MapField(field=mongoengine.IntField(), default={})
    job_imports = mongoengine.ListField(field=mongoengine.StringField(), default=[])
    platform = mongoengine.DictField()
    boot_time = mongoengine.DateTimeField()
    python_version = mongoengine.StringField()
    python_packages = mongoengine.ListField(field=mongoengine.StringField())

    def history(self, offset=0, limit=30, step=0):
        step_filter = {}
        if step and step > 1:
            step_filter = {'index__mod':(step,0)}
        statuses = HostStatus.objects(host=self, **step_filter).order_by('-created')[offset:limit]
        return [s.to_safe_dict(with_host=False) for s in statuses]

    def alive(self):
        recent_count = HostStatus.objects(host=self, created__gte=datetime.utcnow() - timedelta(minutes=0.5)).count()
        return recent_count > 0

    def last_seen_alive(self):
        last_status = HostStatus.objects(host=self).order_by('-created').only('created').first()
        if not last_status or not last_status.created:
            return None
        return last_status.created

    def to_safe_dict(self, alive=False, with_history=False, offset=0, limit=30, step=0):
        r = super(Host, self).to_safe_dict()
        if alive:
            r['alive'] = self.alive()
            r['last_seen_alive'] = self.last_seen_alive()
        if with_history:
            r['history'] = self.history(offset=offset, limit=limit, step=step)
        return r

    def update_status(self):
        partitions = []
        try:
            for f in psutil.disk_partitions():
                usage = psutil.disk_usage(path=f.mountpoint)
                p = {
                    'type': f.fstype,
                    'device': f.device,
                    'mountpoint': f.mountpoint,
                    'total': usage.total,
                    'used': usage.used,
                    'percent': usage.percent,
                }
                partitions.append(p)
        except Exception as e:
            pass

        self_process = psutil.Process(os.getpid())
        processes = [{'ppid': self_process.ppid(), 'pid': self_process.pid, 'cmd': ' '.join(self_process.cmdline())}]
        for c in self_process.children():
            try:
                processes.append({'ppid': c.ppid(), 'pid': c.pid, 'cmd': ' '.join(c.cmdline())})
            except psutil.Error:
                pass

        self.host_status_index += 1

        status = HostStatus()
        status.index = self.host_status_index
        status.host = self
        status.current_jobs = [{'uuid': j.uuid, 'type': j._cls} for j in self.client_service.current_jobs]

        virtual_memory = psutil.virtual_memory()
        swap_memory = psutil.swap_memory()

        status.system_status = {
            'processes': processes,
            'cpu': {
                'percent': psutil.cpu_percent(),
                'percents': psutil.cpu_percent(percpu=True)
            },
            'memory': {
                'virtual': {
                    'total': virtual_memory.total,
                    'used': virtual_memory.used,
                    'percent': virtual_memory.percent,
                },
                'swap': {
                    'total': swap_memory.total,
                    'used': swap_memory.used,
                    'percent': swap_memory.percent,
                },
            },
            'disk': partitions,
            #'disk_io': safe_dict(psutil.disk_io_counters, perdisk=False)
        }
        status.save()

    @classmethod
    def localhost(cls):
        hostname = socket.gethostname()
        hosts = Host.objects(hostname=hostname)
        if not hosts:
            logging.info('Host unknown. Initializing it in the database...')
            host = Host()
            host.hostname = hostname
            host.mac_address = tbx.network.get_mac_address()
            host.platform = tbx.code.safe_dict(platform.uname)
            host.host_status_index = 1
            logging.info("Now, configure Host '%s' through API or Web UI to be able to use it." % hostname)
        else:
            host = hosts[0]
            last_status = HostStatus.objects(host=host).order_by('-created').first()
            if not last_status:
                host.host_status_index = 1
            else:
                host.host_status_index = last_status.index
            logging.info("Host '%s' already found in database." % hostname)

        host.boot_time = datetime.fromtimestamp(psutil.boot_time())
        host.pid = os.getpid()
        host.python_version = sys.version.split(' ')[0]
        host.python_packages = sorted(
            ["%s (%s)" % (i.key, i.version) for i in pip.get_installed_distributions(local_only=False)])
        host.save()
        logging.info("Host '%s' config updated in database." % hostname)
        return host

    def update_slots(self, job_slots=None):
        from jobmanager.common.job import Job, JobTask
        job_classes = tbx.code.get_subclasses(Job)
        job_tasks = tbx.code.get_subclasses(JobTask)
        if not job_slots:
            logging.info('Job Slots not set in env or command line args. Setting to default job defined amount.')
            job_slots = {k.__name__: k.default_slot_amount() for k in job_classes}
        for c in job_classes:
            class_name = c.__name__
            if class_name not in job_slots.keys():
                self.job_slots[class_name] = 0
            else:
                self.job_slots[class_name] = job_slots[class_name]
            logging.info(" - Job type found : %s (%d slots)" % (class_name, self.job_slots[class_name]))
        logging.info("Also found following job tasks : %s" % ', '.join([k.__name__ for k in job_tasks]))
        return self.save()

    def check_capacity(self):
        if self.job_slots:
            #logging.info("Jobs types allowed : %s" % (', '.join(self.job_slots.keys())))
            logging.info("Jobs capacity :")
            total_capacity = 0
            for c in self.job_slots:
                logging.info(" - %s\t: %d" % (c, self.job_slots[c]))
                total_capacity += self.job_slots[c]

            if total_capacity == 0:
                logging.error("No job capacities setup. Configure host to add slots to some job types.")
                raise common.ConfigurationException(
                    "No job capacities setup. Configure host to add slots to some job types (see --add-slot option).")
        else:
            logging.error(
                "No job class found to be run. Configure host to import packages that contain job subclasses.")
            raise common.ConfigurationException(
                "No Job sub-class found in imports. Please configure host (see --import option).")

    def do_import(self, imports):
        return common.safely_import_from_name(imports)

    @classmethod
    def get_all_alive(cls):
        raise NotImplementedError()


class HostStatus(common.BaseDocument):

    meta = {
        'ordering': ['-created'],
        'max_documents': 200000,
        'max_size': 200000000,
        'queryset_class': common.SerializableQuerySet,
        'indexes': [
            'created',
            'host'
        ]
    }
    host = mongoengine.CachedReferenceField(Host, fields=['hostname'], reverse_delete_rule=mongoengine.CASCADE)
    index = mongoengine.LongField(required=True, default=0)
    system_status = mongoengine.DictField(default={})
    current_jobs = mongoengine.ListField(field=mongoengine.DictField(), default=[])
    updated = None

    def to_safe_dict(self, with_host=True):
        r = super(HostStatus, self).to_safe_dict()
        if not with_host:
            del r['host']
            del r['type']
        return r



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu
"""
(c) 2015 Ronan Delacroix
Job Manager Job Abstract Class
:author: Ronan Delacroix
"""
import os
import logging
import mongoengine
import mongoengine.signals
import tbx
import tbx.process
import tbx.service
import tbx.log
import tbx.text
import uuid as UUID
import traceback
import tempfile
import shutil
from datetime import datetime, timedelta
import jobmanager.common as common
from .host import Host
from tbx.code import cached_property


job_status_to_icon = {
    'new': "\u23F8",
    'pending': "\u23F8",
    'running': "\u25B6",
    'success': "\u2605",
    'error': "\u2716"
}


class Job(common.NamedDocument, common.Runnable, common.LogProxy, common.AutoDocumentable):

    meta = {
        'collection': 'jobs',
        'indexes': [
            'status',
            'created',
        ]
    }

    status_text = mongoengine.StringField(required=True, default="")
    hostname = mongoengine.StringField()
    completion = mongoengine.IntField(required=True, min_value=0, max_value=100, default=0)
    timeout = mongoengine.IntField(min_value=0, default=43200)  # 12 hours
    ttl = mongoengine.IntField(min_value=1, default=1)
    history = mongoengine.ListField(field=mongoengine.DictField(), default=[])

    def __str__(self):
        return "%s %s" % (self.name, job_status_to_icon.get(self.status, self.status))

    @classmethod
    def default_slot_amount(cls):
        """
        Returns the default amount of job that can be run at the same time on the same machine/client.
        Override and set to math.inf for no limiting amount.
        You can base that how much CPU cores you have or anything else.
        Default is 1 job at a time by default.
        """
        return 1

    @cached_property
    def extra_log_arguments(self):
        return {
            'job_type': self.__class__.__name__,
            'job_uuid': self.uuid,
            'job_status': self.status,
        }

    def update_status(self, completion=None, text=None):
        if text:
            self.status_text = text

        if completion:
            self.completion = completion

        log = self.log_info
        if self.status == 'error':
            log = self.log_error

        log("Progress update : {progress:5.1f}% - {message}".format(
            progress=self.completion,
            message=self.status_text
        ))

        self.update(
            add_to_set__history={'t': datetime.utcnow(), 'm': self.status_text, 'c': self.completion, 's': self.status},
            status=self.status,
            details=self.details,
            completion=self.completion,
            status_text=self.status_text,
            started=self.started,
            finished=self.finished
        )

    def update_progress(self, completion, text=None):
        self.update_status(completion=completion, text=text)

    def save_as_successful(self, text="Job Successful"):
        self.update_status(100, text=text)

    def save_as_error(self, text="Job Error"):
        self.update_status(text=text)


mongoengine.signals.pre_save.connect(common.update_modified)


class JobTask(mongoengine.EmbeddedDocument, common.Runnable, common.LogProxy, common.AutoDocumentable):
    meta = {
        'abstract': True,
    }

    status = mongoengine.StringField(required=True, default="pending",
                                     choices=('new', 'pending', 'running', 'success', 'error'))
    details = mongoengine.StringField(required=False)

    @property
    def job(self):
        if isinstance(self._instance, JobTask):
            return self._instance.job
        return self._instance

    @cached_property
    def extra_log_arguments(self):
        extra_log_arguments = {}
        if isinstance(self.job, common.LogProxy):
            extra_log_arguments = self.job.extra_log_arguments
        extra_log_arguments['task'] = self.name
        return extra_log_arguments

    def __str__(self):
        return "%s > %s" % (self.job, self.name)

    def get_hash(self):
        import base64
        import hashlib
        return base64.b64encode(
            hashlib.sha1(mongoengine.EmbeddedDocument.to_json(self, sort_keys=True).encode()).digest()).decode().strip('=').replace("+", "-")

    def update_status(self, completion=None, text=None):
        #TODO : Review this part // Completion between tasks and jobs is not clear.
        if text:
            self.job.status_text = text

        if completion:
            self.job.completion = completion

        log = self.log_info
        if self.status == 'error':
            log = self.log_error

        log("Progress update : {progress:5.1f}% - {message}".format(
            progress=self.job.completion,
            message=text
        ))

    def update_progress(self, completion, text=None):
        self.update_status(completion=completion, text=text)


def make_job(job_name, **kwargs):
    """
    Decorator to create a Job from a function.
    Give a job name and add extra fields to the job.

        @make_job("ExecuteDecJob",
                  command=mongoengine.StringField(required=True),
                  output=mongoengine.StringField(default=None))
        def execute(job: Job):
            job.log_info('ExecuteJob %s - Executing command...' % job.uuid)
            result = subprocess.run(job.command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            job.output = result.stdout.decode('utf-8') + " " + result.stderr.decode('utf-8')

    """
    def wraps(func):
        kwargs['process'] = func
        job = type(job_name, (Job,), kwargs)
        globals()[job_name] = job
        return job
    return wraps



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu
"""
:Author: Ronan Delacroix
:Copyright: (c) 2018 Ronan Delacroix
"""
import mongoengine
import jobmanager.common


class DockerImage(jobmanager.common.NamedDocument):
    meta = {
        'queryset_class': jobmanager.common.SerializableQuerySet,
        'indexes': [
            'uuid',
            'created',
            'name'
        ]
    }
    name = mongoengine.StringField(required=True)
    image_id = mongoengine.StringField()
    url = mongoengine.ListField(mongoengine.StringField())
    tags = mongoengine.ListField(field=mongoengine.StringField())
    jobs = mongoengine.ListField(field=mongoengine.StringField())
    tasks = mongoengine.ListField(field=mongoengine.StringField())
    requirements = mongoengine.ListField(field=mongoengine.StringField())
    apt_packages = mongoengine.ListField(field=mongoengine.StringField())
    dockerfile = mongoengine.StringField()
